from __future__ import annotations

from numbers import Number

import jijmodeling as jm
import numpy as np

import dataclasses
from jijzept.entity.schema import SolverType
from jijzept.response import JijModelingResponse
from jijzept.sampler.base_sampler import (
    JijZeptBaseSampler,
    ParameterSearchParameters,
    check_kwargs_against_dataclass,
    merge_params_and_kwargs,
    sample_model,
)


@dataclasses.dataclass
class JijSQBMParameters:
    """Manage Parameters for using Toshiba Simulated Bifurcation Machine.

    An parameter with none will be set to the default value in Simulated Bifurcation Machine.
    For more information on the parameters. Please see [this page](https://learn.microsoft.com/ja-jp/azure/quantum/provider-toshiba).

    Attributes:
        steps (Optional[PositiveInt]): Number of steps.
        loops (Optional[PositiveInt]): Number of loops.
        timeout (Optional[PositiveFloat]): Timeout.
        target (Optional[PositiveFloat]): Target.
        maxout (Optional[PositiveInt]): Maxout.
        dt (Optional[PositiveFloat]): dt.
        C (Optional[PositiveFloat]): C.
        algo (Optional[str]): Algorithm.
        auto (Optional[bool]): Auto.
    """

    steps: int | None = None
    loops: int | None = None
    timeout: float | None = None
    target: float | None = None
    maxout: int | None = None
    dt: float | None = None
    C: float | None = None
    algo: str | None = None
    auto: bool | None = None


class JijSQBMSampler(JijZeptBaseSampler):
    solver_type = SolverType(queue_name="qiosolver", solver="SQBM+")
    jijmodeling_solver_type = SolverType(
        queue_name="qiosolver", solver="SQBM+ParaSearch"
    )

    def sample_model(
        self,
        model: jm.Problem,
        feed_dict: dict[str, Number | list | np.ndarray],
        multipliers: dict[str, Number] | None = None,
        fixed_variables: dict[str, dict[tuple[int, ...], int | float]] | None = None,
        needs_square_dict: dict[str, bool] | None = None,
        search: bool = False,
        num_search: int = 15,
        algorithm: str | None = None,
        parameters: JijSQBMParameters | None = None,
        max_wait_time: int | float | None = None,
        sync: bool = True,
        queue_name: str | None = None,
        **kwargs,
    ) -> JijModelingResponse:
        """Sample using JijModeling by means of Microsoft QIO Simulated Annealing solver.

        To configure the solver, instantiate the `JijQIOXXParameters` class for your desired solver and pass the instance to the `parameters` argument.

        Args:
            model (jijmodeling.Problem): Mathematical model is constracted by JijModeling.
            feed_dict (Dict[str, Union[Number, list, np.ndarray]]): The actual values to be assigned to the placeholders.
            multipliers (Dict[str, Number], optional): The actual multipliers for penalty terms, derived from constraint conditions.
            fixed_variables (Dict[str, Dict[Tuple[int, ...], Union[int, float]]]): dictionary of variables to fix.
            needs_square_dict (Dict[str, bool], optional): If `True`, the corresponding constraint is squared when added as a penalty to the QUBO. When the constraint label is not added to the 'needs_square_dict', it defaults to `True` for linear constraints and `False` for non-linear constraints.
            search (bool, optional): If `True`, the parameter search will be carried out, which tries to find better values of multipliers for penalty terms.
            num_search (int, optional): The number of parameter search iteration. Defaults to set 15. This option works if `search` is `True`.
            algorithm (Optional[str]): Algorithm for parameter search. Defaults to None.
            parameters (JijSQBMParameters | None, optional): parameters for SQBM+. Defaults to None.
            max_wait_time (int | float | None, optional): The number of timeout [sec] for post request. If `None`, 60 (one minute) will be set. Please note that this argument is for the `jijzept` timeout and not for configuring solver settings, such as solving time.
            sync (bool, optional): Synchronous mode.
            queue_name (Optional[str], optional): Queue name.
            **kwargs: QIO parameters using **kwargs. If both `**kwargs` and `parameters` are exist,　the value of `**kwargs` takes precedence.

        Returns:
            JijModelingResponse: Stores minimum energy samples and other information.

        Examples:
            ```python
            import jijzept as jz
            import jijmodeling as jm
            n = jm.Placeholder('n')
            x = jm.BinaryVar('x', shape=(n,))
            d = jm.Placeholder('d', ndim=n)
            i = jm.Element("i", belong_to=n)
            problem = jm.Problem('problem')
            problem += jm.Sum(i, d[i] * x[i])
            sampler = jz.JijSQBMSampler(config='config.toml')
            response = sampler.sample_model(problem, feed_dict={'n': 5, 'd': [1,2,3,4,5]})
            ```
        """

        if multipliers is None:
            multipliers = {}
        if fixed_variables is None:
            fixed_variables = {}

        if needs_square_dict is not None:
            warnings.warn(
                message="The argument `needs_square_dict` is deprecated. Please don't use it.",
                stacklevel=2,
            )

        check_kwargs_against_dataclass(kwargs, JijSQBMParameters)
        param_dict = merge_params_and_kwargs(parameters, kwargs, JijSQBMParameters)

        if queue_name is None:
            queue_name = self.jijmodeling_solver_type.queue_name

        para_search_params = ParameterSearchParameters(
            multipliers=multipliers,
            mul_search=search,
            num_search=num_search,
            algorithm=algorithm,
        )

        sample_set = sample_model(
            self.client,
            self.jijmodeling_solver_type.solver,
            queue_name=queue_name,
            problem=model,
            instance_data=feed_dict,
            fixed_variables=fixed_variables,
            parameter_search_parameters=para_search_params,
            max_wait_time=max_wait_time,
            sync=sync,
            **param_dict,
        )
        return sample_set
